#include "dxstdafx.h"
#include ".\GameManager.h"
#include "fmod.h"
#include "Object3d.h"
#include "SoundNode.h"

GameManager::GameManager(void)
{
	this->consoleStarted = false;
	this->activeScene=0;

	//create ResourceManager
	this->resManager.setGameManager(this);
	this->firstFrame = true;
	this->finalEffect = NULL;
	this->shadowMapEffect = NULL;
	this->gs.setSceneAlpha(1);

	for(int i=0;i<this->NB_EFFECTS;i++) {
		this->effectList[i] = NULL;
		this->effectReleaseCount[i] = 0;
	}
}

GameManager::~GameManager(void)
{
	SAFE_RELEASE(this->finalRenderTarget);

	for(int i=0;i<this->NB_EFFECTS;i++) {
		if(this->effectList[i]) {
			SAFE_RELEASE(this->effectList [i]);
		}
	}
}

void GameManager::setScreenDimension(int width, int height)
{
	this->screenWidth = width;
	this->screenHeight = height;

	this->screenResolution[0] = width;
	this->screenResolution[1] = height;
}

void GameManager::keyPressed(UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext)
{
	if( bKeyDown )
    {
		switch( nChar )
			{
				case VK_ESCAPE:
					break;
				case VK_F1:
					this->gs.switchGUI();
					break;
				case VK_F2:
					break;
				case VK_F3:
					break;
				case VK_F4:
					break;
				case VK_F5:
					break;
				case VK_F6:
					break;
				case VK_F7:
					break;
				case VK_F8:
					break;
				case VK_F9:
					break;
				case VK_F11:
					break;
			}
	}
	this->activeScene->setKeyPressed(nChar, bKeyDown, bAltDown);
}

void GameManager::setMouseStatus(bool bLeftButtonDown,  bool bRightButtonDown,bool bMiddleButtonDown,
							  bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta,
							  int xPos, int yPos)
{
	this->activeScene->setMouseStatus(bLeftButtonDown, bRightButtonDown, bMiddleButtonDown, bSideButton1Down, bSideButton2Down, nMouseWheelDelta, xPos, yPos);
}

void GameManager::addScene(Scene &newScene)
{
	if(this->activeScene==0) {
		this->activeScene = &newScene;
	}
	this->sceneVector.push_back(&newScene);
}

void GameManager::removeScene(Scene &oldScene)
{
	std::vector<Scene*>::iterator it;
	
	bool found = false;
	int index = 0;
	for(it=this->sceneVector.begin();it!=this->sceneVector.end();it++) {
		if(&oldScene == *it) {
			found = true;
			break;
		}
		index++;
	}
}

void GameManager::setActiveScene(Scene &aScene)
{
	this->activeScene = &aScene;
}

void GameManager::printToConsole(std::string output) 
{
	#if defined(DEBUG) | defined(_DEBUG)
	if (!consoleStarted) {
        AllocConsole();
		outputHandle = GetStdHandle(STD_OUTPUT_HANDLE);
		consoleStarted = true;
	}
	output += "\n";
	DWORD dwCharsWritten;
	WriteConsoleA(outputHandle, output.c_str(), (DWORD)output.length(), &dwCharsWritten, NULL);
	#endif
}

void GameManager::initGame() {
	gs.initScene(*this);
	gs.setVisible(true);
	this->addScene(gs);
	this->setActiveScene(gs);
}

void GameManager::updateGame(float fElapsedTime) 
{
	FSOUND_Update();
	
	if(this->firstFrame) {
		this->firstFrame = false;
	}
	this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	this->device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(0, 0, 0, 0), 1.0f, 0);
	
	UINT numberOfScenes = (UINT)this->sceneVector.size();
	for(UINT i = 0; i < numberOfScenes; i++) {
		this->sceneVector.at(i)->renderScene(fElapsedTime);
	}

	this->device->SetRenderTarget(0, this->finalRenderTarget);
	this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	this->device->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);
	this->device->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
	this->device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_DESTALPHA);
	this->device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);

	for(UINT i = 0; i < numberOfScenes; i++) {
		if(this->sceneVector.at(i)->isVisible()) {
			this->sceneVector.at(i)->renderFinalImage(this->finalEffect);
		}
	}

	this->device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
	this->device->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
}


void GameManager::OnLostDevice( void* pUserContext )
{
	SAFE_RELEASE(this->finalRenderTarget);
	this->resManager.OnLostDevice();
	std::vector<Scene *>::iterator it;
	for(it=this->sceneVector.begin();it!=this->sceneVector.end();it++) {
		(*it)->OnLostDevice(pUserContext);
	}
}

void GameManager::OnDestroyDevice( void* pUserContext )
{
	this->resManager.OnDestroyDevice();
	
	this->releaseEffect(this->finalEffect);
	this->finalEffect = NULL;
	
	this->releaseEffect(this->shadowMapEffect);
	this->shadowMapEffect = NULL;

	std::vector<Scene *>::iterator it;
	for(it=this->sceneVector.begin();it!=this->sceneVector.end();it++) {
		(*it)->OnDestroyDevice(pUserContext);
	}

	for(int i=0;i<this->NB_EFFECTS;i++) {
		if(this->effectList[i] != NULL) {
			this->printToConsole("WARNING! Not all Effects released!!");
		}
	}
}

HRESULT GameManager::OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext ) 
{
	this->device = pd3dDevice;
	return S_OK;
}


HRESULT GameManager::OnResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
	this->screenWidth = pBackBufferSurfaceDesc->Width;
	this->screenHeight = pBackBufferSurfaceDesc->Height;

	this->device = pd3dDevice;
	this->device->GetRenderTarget(0, &this->finalRenderTarget);
	this->finalEffect = this->loadEffect(GameManager::EFFECT_FINAL, L"./shaders/finalRender.fx");
	this->shadowMapEffect = this->loadEffect(GameManager::EFFECT_SHADOWMAP, L"./shaders/shadowMap.fx");
	this->depthMapEffect = this->loadEffect(GameManager::EFFECT_DEPTHMAP, L"./shaders/depthPass.fx");
	this->simpleMeshEffect = this->loadEffect(GameManager::EFFECT_SIMPLEMESH, L"./shaders/simpleMesh.fx");
	
	std::vector<Scene *>::iterator it;
	for(it=this->sceneVector.begin();it!=this->sceneVector.end();it++) {
		(*it)->OnResetDevice(pd3dDevice, pBackBufferSurfaceDesc, pUserContext );
	}
	
	return S_OK;
}

ID3DXEffect* GameManager::getEffect(UINT id)
{
	if(id<0 || id>=this->NB_EFFECTS)
		return NULL;
	ID3DXEffect* effect = this->effectList[id];
	if(effect) {
		this->effectReleaseCount[id]++;
	}
	return effect;
}

void GameManager::releaseEffect(ID3DXEffect* effect)
{
}

ID3DXEffect* GameManager::loadEffect(UINT id, LPCWSTR filename) {
	ID3DXEffect* effect;

	if(!this->effectList[id]) {
		//DWORD dwShaderFlags = D3DXSHADER_PREFER_FLOW_CONTROL | D3DXSHADER_SKIPOPTIMIZATION ;
		DWORD dwShaderFlags = D3DXSHADER_PREFER_FLOW_CONTROL | D3DXSHADER_SKIPOPTIMIZATION | D3DXSHADER_USE_LEGACY_D3DX9_31_DLL;
		#ifdef DEBUG_VS
			dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
		#endif
		#ifdef DEBUG_PS
			dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
		#endif

		// Read the D3DX effect file
		ID3DXBuffer* errBuff = NULL;
		if (FAILED(D3DXCreateEffectFromFile( DXUTGetD3DDevice(),filename, NULL, NULL, dwShaderFlags,
			NULL, &effect, &errBuff )))
		{
			int BufSize = errBuff->GetBufferSize();

			// displaying error message of arbitrary length
			wchar_t* wbuf = new wchar_t[BufSize];
			size_t numConverted;
			mbstowcs_s(&numConverted, wbuf, BufSize, (const char*)errBuff->GetBufferPointer(), BufSize );
			MessageBox(NULL, wbuf, L".fx Compilation Error", MB_ICONERROR);		// show error message

			delete wbuf;
			exit(-1);
		} else {
			this->effectList[id] = effect;
			this->effectReleaseCount[id]++;
		}
		return effect;
	} else {
		this->effectReleaseCount[id]++;
		return this->effectList[id];
	}
}

void GameManager::OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
	this->activeScene->OnGUIEvent(nEvent, nControlID, pControl, pUserContext);
}
