#include "dxstdafx.h"
#include "LowMemShadowPass.h"
#include "LowMemLight.h"
#include "Camera.h"
#include "Scene.h"
#include "GameManager.h"

LowMemShadowPass::LowMemShadowPass(void)
{
	light = NULL;

	left.texture = NULL;
	left.surface = NULL;
	right.texture = NULL;
	right.surface = NULL;
	left.shaded = false;
	right.shaded = false;

	tempTexture = NULL;
	tempSurface = NULL;
	maxLevel = 8;
	smartReprojection = false;
}

LowMemShadowPass::~LowMemShadowPass(void)
{
	if(tempSurface) SAFE_RELEASE(tempSurface);
	if(tempTexture) SAFE_RELEASE(tempTexture);

	std::vector<depthMapNode>::iterator it;
	for(it=lowResMapList.begin();it!=lowResMapList.end();it++) {
		SAFE_RELEASE(it->surface);
		SAFE_RELEASE(it->texture);
	}
}

void LowMemShadowPass::setActiveLight(LowMemLight* _light)
{
	light = _light;
	this->setProjectionMatrix(*scene->getActiveCamera()->getProjectionMatrix());

	int mapRes = light->getShadowMapSize();
	errorSurface = light->getErrorSurface();

	if(tempSurface) SAFE_RELEASE(tempSurface);
	if(tempTexture) SAFE_RELEASE(tempTexture);

	HRESULT hr;	
	V( DXUTGetD3DDevice()->CreateTexture(mapRes, mapRes, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &tempTexture, NULL ) );
	V( tempTexture->GetSurfaceLevel(0, &tempSurface) );

	lowRes = light->getLowResShadowMapSize();

	createDepthMap(left);
	createDepthMap(right);
}

void LowMemShadowPass::executePass()
{
	this->resetLowResMaps();
	this->clearSumTexture(false);

	if(this->scene->shadowingMode != Scene::SHMODE_BIGPCF) {
		left.deltaX =  -light->getLightDimension()*0.5f;
		right.deltaX = light->getLightDimension()*0.5f;

		calcMatrices(left);
		calcMatrices(right);

		nbMaps = 0;
		refineShadow(left, right, 0);

		light->swapLightSumMaps();
	} else {
		//Make one ShadowMap and use big pcf filter
		left.deltaX = 0;
		calcMatrices(left);
		nbMaps = 0;
		bigPCFShadow(left);
		light->swapLightSumMaps();
	}
}

void LowMemShadowPass::calcMatrices(depthMapNode& node)
{

	D3DXMATRIX tempWorldMat;
	D3DXMATRIX translationMat;
	D3DXMatrixTranslation(&translationMat, node.deltaX, 0, 0);

	//Render First DepthMap
	D3DXMatrixMultiply(&tempWorldMat, light->getWorldMatrix(), &translationMat);
	node.worldMat = tempWorldMat;
}

void LowMemShadowPass::refineShadow(depthMapNode leftNode, depthMapNode rightNode, int level)
{
	depthMapNode middle;

	if(level>=maxLevel) {
		releaseLowResMap(leftNode);
		return;
	}
	level++;

	//Shade and reproject if necessary
	if(this->smartReprojection) {
		if(!leftNode.shaded) shadeAndReproject(leftNode, false, leftNode.worldMat);
		if(!rightNode.shaded) shadeAndReproject(rightNode, true, leftNode.worldMat);
	} else {
		if(!leftNode.shaded) shadeAndReproject(leftNode, true, *light->getWorldMatrix());
		if(!rightNode.shaded) shadeAndReproject(rightNode, true, *light->getWorldMatrix());
	}

	if(!this->isQualityOkay(leftNode, rightNode)) {
		this->createDepthMap(middle);
		middle.deltaX = (rightNode.deltaX + leftNode.deltaX)*0.5f;
		this->calcMatrices(middle);
		
		refineShadow(leftNode, middle, level);		
		middle.shaded = true;
		refineShadow(middle, rightNode, level);
	} else {
		releaseLowResMap(leftNode);
	}
}

void LowMemShadowPass::shadeAndReproject(depthMapNode& node, bool reproject, D3DXMATRIX refMatrix)
{
	HRESULT hr;

	//Calculate DepthMap
	this->setViewMatrix(node.worldMat);
	this->calculateMatrices();
	
	if(this->scene->viewPointMode == Scene::VIEWPOINT_CAMERA || reproject) {
		this->setRenderTarget(tempSurface);
	} else {
		this->setRenderTarget(node.surface);
	}

	this->renderPassId = Scene::PASS_DEPTH;
	RenderPass::executePass();

	//Set current Texture, LightWorldMatrix and Number of ShadowMaps
	if(this->scene->viewPointMode == Scene::VIEWPOINT_CAMERA || reproject) {
		light->setCurrentDepthMap(tempTexture);
	} else {
		light->setCurrentDepthMap(node.texture);
	}

	light->setCurrentWorldMatrix(&node.worldMat);
	light->setNbShadowMaps(++nbMaps);

	//Calculate Shadow from selected View
	this->setViewMatrix(*(scene->getActiveCamera()->getViewMatrix()));
	this->calculateMatrices();
	light->swapLightSumMaps();
	
	if(this->scene->sumMode == scene->SUM_TWOPASS) {
		this->setRenderTarget(light->getThirdRenderTarget());
	} else {
		this->setRenderTarget(light->getLightSumRenderTarget());
	}

	this->renderPassId = Scene::PASS_ADDSHADOWMAP;
	RenderPass::executePass();

	UINT cPass;
	ID3DXEffect* shadowEffect;
	shadowEffect = this->scene->manager->getEffect(GameManager::EFFECT_SHADOWMAP);

	if(scene->sumMode == scene->SUM_TWOPASS) {
		scene->device->SetRenderTarget(0, light->getLightSumRenderTarget());

		//set the 2 textures here...
		V( shadowEffect->SetTexture("g_txShadowMap", light->getThirdTexture()) );
		V( shadowEffect->SetTexture("g_txSourceMap", light->getLightSumTexture()) );

		V( shadowEffect->SetTechnique("SumAfterwards") );
		
		V( shadowEffect->CommitChanges() );

		V( shadowEffect->Begin( &cPass, 0 ) );
		V( shadowEffect->BeginPass( 0 ) );
		V( this->device->BeginScene() );
		
		V( this->scene->getQuadMesh()->DrawSubset(0) );

		//V( this->device->SetTexture(0, 0) );
		V( this->device->EndScene() );
		V( shadowEffect->EndPass() );
		V( shadowEffect->End() );
	}

	//Do reprojection
	if(this->scene->viewPointMode == Scene::VIEWPOINT_CAMERA || reproject) {
		this->device->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
		this->device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);

		//Rendertarget auf depthMap setzen
		this->setRenderTarget(node.surface);
		this->setViewMatrix(node.worldMat);
		scene->setLightViewMatrix(node.worldMat);
		if(this->scene->viewPointMode == Scene::VIEWPOINT_LIGHT) {
			scene->setLightReferenceViewMatrix(refMatrix);
		} else {
			scene->setLightReferenceViewMatrix(*scene->getActiveCamera()->getViewMatrix());
		}

		this->calculateMatrices();
		V( shadowEffect->SetTexture("g_txSourceMap", tempTexture) );
		V( shadowEffect->SetTechnique("Reprojection") );
		V( shadowEffect->CommitChanges() );

		this->renderPassId = Scene::PASS_REPROJECTION;

		RenderPass::executePass();

		this->renderPassId = Scene::PASS_SOFTSHADOWMAP;
		light->setCurrentDepthMap(node.texture);
	}

	this->setRenderTarget(node.surface);
	this->setViewMatrix(node.worldMat);
	scene->setLightViewMatrix(node.worldMat);

	this->renderPassId = Scene::PASS_SOFTSHADOWMAP;
	light->setCurrentDepthMap(node.texture);

	node.shaded = true;
}

void LowMemShadowPass::bigPCFShadow(depthMapNode node)
{
	HRESULT hr;

	//Calculate DepthMap
	this->setViewMatrix(node.worldMat);
	this->calculateMatrices();

	this->setRenderTarget(node.surface);

	this->renderPassId = Scene::PASS_DEPTH;
	RenderPass::executePass();

	//Set current Texture, LightWorldMatrix and Number of ShadowMaps
	light->setCurrentDepthMap(node.texture);

	light->setCurrentWorldMatrix(&node.worldMat);
	light->setNbShadowMaps(++nbMaps);

	//Calculate Shadow from selected View
	this->setViewMatrix(*(scene->getActiveCamera()->getViewMatrix()));
	
	this->calculateMatrices();
	light->swapLightSumMaps();
	
	if(this->scene->sumMode == scene->SUM_TWOPASS) {
		this->setRenderTarget(light->getThirdRenderTarget());
	} else {
		this->setRenderTarget(light->getLightSumRenderTarget());
	}

	this->renderPassId = Scene::PASS_ADDSHADOWMAP;
	RenderPass::executePass();

	UINT cPass;
	ID3DXEffect* shadowEffect;
	shadowEffect = this->scene->manager->getEffect(GameManager::EFFECT_SHADOWMAP);

	if(scene->sumMode == scene->SUM_TWOPASS) {
		scene->device->SetRenderTarget(0, light->getLightSumRenderTarget());

		//set the 2 textures here...
		V( shadowEffect->SetTexture("g_txShadowMap", light->getThirdTexture()) );
		V( shadowEffect->SetTexture("g_txSourceMap", light->getLightSumTexture()) );

		V( shadowEffect->SetTechnique("SumAfterwards") );
		
		V( shadowEffect->CommitChanges() );

		V( shadowEffect->Begin( &cPass, 0 ) );
		V( shadowEffect->BeginPass( 0 ) );
		V( this->device->BeginScene() );
		
		V( this->scene->getQuadMesh()->DrawSubset(0) );

		V( this->device->EndScene() );
		V( shadowEffect->EndPass() );
		V( shadowEffect->End() );
	}
}


void LowMemShadowPass::createDepthMap(depthMapNode& node)
{
	HRESULT hr;
	int i=1;
	std::vector<depthMapNode>::iterator it;
	for(it=lowResMapList.begin();it!=lowResMapList.end();it++) {
		if(!it->inUse) {
			node.texture = it->texture;
			node.surface = it->surface;
			node.inUse = true;
			node.shaded = false;
			it->inUse = true;
			it->shaded = false;
			this->maxUsedMaps = max(i, this->maxUsedMaps);
			return;
		}
		i++;
	}

	V( DXUTGetD3DDevice()->CreateTexture(lowRes, lowRes, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &node.texture, NULL ) );
	V( node.texture->GetSurfaceLevel(0, &node.surface) );
	
	node.inUse = true;
	node.shaded = false;
	lowResMapList.push_back(node);
	this->maxUsedMaps++;
}

void LowMemShadowPass::releaseLowResMap(depthMapNode node)
{
	if(node.texture == left.texture || node.texture == right.texture) {
		return;
	}
	std::vector<depthMapNode>::iterator it;
	for(it=lowResMapList.begin();it!=lowResMapList.end();it++) {
		if(node.texture == it->texture) {
			it->inUse = false;
		}
	}
}

void LowMemShadowPass::resetLowResMaps()
{
	this->maxUsedMaps = 0;
	std::vector<depthMapNode>::iterator it;
	for(it=lowResMapList.begin();it!=lowResMapList.end();it++) {
		if(it->texture != left.texture && it->texture != right.texture) {
			it->inUse = false;
		}
	}
}

bool LowMemShadowPass::isQualityOkay(depthMapNode leftNode, depthMapNode rightNode)
{
	if(this->scene->shadowingMode == Scene::SHMODE_SIMPLE) {
		return false;
	}

	float lightDim = light->getLightDimension();

	//OcclusionQuery
	HRESULT hr = this->scene->device->CreateQuery(D3DQUERYTYPE_OCCLUSION, NULL);
	IDirect3DQuery9* query;
	DWORD numberOfPixelsDrawn;
	this->scene->device->CreateQuery(D3DQUERYTYPE_OCCLUSION, &query);
	query->Issue(D3DISSUE_BEGIN);
		
	this->calcDifference(leftNode.texture, rightNode.texture);	
	
	// Add an end marker to the command buffer queue.
	query->Issue(D3DISSUE_END);

	// Force the driver to execute the commands from the command buffer.
	// Empty the command buffer and wait until the GPU is idle.
	while(S_FALSE == query->GetData((void*) &numberOfPixelsDrawn, sizeof(DWORD), D3DGETDATA_FLUSH ));

	SAFE_RELEASE(query);
	return light->isErrorOkay(numberOfPixelsDrawn);
}

void LowMemShadowPass::calcDifference(IDirect3DTexture9* leftMap, IDirect3DTexture9* rightMap)
{
	HRESULT hr;
	UINT cPass;
	ID3DXEffect* shadowEffect;
	shadowEffect = this->scene->manager->getEffect(GameManager::EFFECT_SHADOWMAP);


	V( this->scene->device->SetRenderTarget(0, this->light->getErrorSurface()) );
	DXUTGetD3DDevice()->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(255, 0, 0, 255), 0.8f, 0);

	V( shadowEffect->SetTechnique("CompareDepthMaps") );
	V( shadowEffect->SetTexture("g_txLeftMap", leftMap) );
	V( shadowEffect->SetTexture("g_txRightMap", rightMap) );
	V( shadowEffect->SetFloat("errorThreshold", light->getErrorThreshold()) );
	V( shadowEffect->CommitChanges() );

	V( shadowEffect->Begin( &cPass, 0 ) );
	V( shadowEffect->BeginPass( 0 ) );
	V( this->device->BeginScene() );
	
	V( this->scene->getQuadMesh()->DrawSubset(0) );

	V( this->device->EndScene() );
	V( shadowEffect->EndPass() );
	V( shadowEffect->End() );
}

int LowMemShadowPass::getNbDepthMaps()
{
	return (int)this->lowResMapList.size();
}

IDirect3DTexture9* LowMemShadowPass::getDepthMapAt(int id)
{
	return this->lowResMapList.at(id).texture;
}

void LowMemShadowPass::clearSumTexture(bool onlyDepth)
{
	this->scene->device->SetRenderTarget(0, light->getLightSumRenderTarget());
	if(!onlyDepth) {
		DXUTGetD3DDevice()->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(255, 0, 0, 0), 1.0f, 0);
	} else {
		DXUTGetD3DDevice()->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(255, 0, 0, 0), 1.0f, 0);
	}
	this->light->swapLightSumMaps();
	if(!onlyDepth) {
		DXUTGetD3DDevice()->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(255, 0, 0, 0), 1.0f, 0);
	} else {
		DXUTGetD3DDevice()->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(255, 0, 0, 0), 1.0f, 0);
	}
	this->light->swapLightSumMaps();
}

