#pragma once
#include "renderpass.h"
#include <vector>

class LowMemLight;

/**
LowMemShadowPass class
*/

class LowMemShadowPass :
	public RenderPass
{
public:
	/**
	Constructor of the SoftShadow Renderpass
	*/
	LowMemShadowPass(void);
	/**
	Destructor of the SoftShadow Renderpass
	*/
	~LowMemShadowPass(void);

	/**
	Starts the calculation of the softshadow
	*/
	virtual void executePass();
	/**
	Activates the current light
	*/
	void setActiveLight(LowMemLight* _light);
	/**
	Returns the number of depthMaps
	*/
	int getNbDepthMaps();
	/**
	Returns depthMap with given id
	*/
	IDirect3DTexture9* getDepthMapAt(int id);

	/**
	Structure for each depthMap that
	stores the texture, surface, the
	coordinates from which they where taken, the offset deltaX
	and two flags for the algorithm managment
	*/
	struct depthMapNode {
		IDirect3DTexture9* texture;
		IDirect3DSurface9* surface;
		D3DXMATRIX worldMat;
		float deltaX;
		bool shaded;
		bool inUse;
	};

	/**
	Activate smartReprojection to reduce the number of depthMaps by one.
	Be aware: This is a feature in alpha state!
	*/
	bool smartReprojection;

	/**
	Number of maximum used depthMaps
	*/
	int maxUsedMaps;
protected:
	LowMemLight* light;
	int nbMaps;
	int lowRes;
	int maxLevel;

	depthMapNode left;
	depthMapNode right;

	IDirect3DTexture9* tempTexture;
	IDirect3DSurface9* tempSurface;
	IDirect3DSurface9* errorSurface;

	std::vector<depthMapNode> lowResMapList;

	/**
	Releases a depthMap
	*/
	void releaseLowResMap(depthMapNode node);
	/**
	Resets all depthMaps
	*/
	void resetLowResMaps();
	/**
	clears the Sum Textures
	*/
	void clearSumTexture(bool onlyDepth);
	/**
	Creates or reuses a available depthMap
	*/
	void createDepthMap(depthMapNode& node);

	/**
	Subdivides the space between two shadow maps (by generating a new shadowmap)
	*/
	void refineShadow(depthMapNode leftNode, depthMapNode rightNode, int level);
	/**
	Performes the shadowshading and the previous reprojection to check against the error bound
	*/
	void shadeAndReproject(depthMapNode& node, bool reproject, D3DXMATRIX refMatrix);
	/**
	Calculates the needed matrices for a given depthMap node
	*/
	void calcMatrices(depthMapNode& node);

	/**
	Uses occlusion queries to calculate differences between two shadowmaps and
	returns if a refinement is needed
	*/
	bool isQualityOkay(depthMapNode leftNode, depthMapNode rightNode);

	/**
	Performs the rendering code for difference calculation
	*/
	void calcDifference(IDirect3DTexture9* leftMap, IDirect3DTexture9* rightMap);

	/**
	Renders the scene with one shadowmap using a big PCF kernel
	*/
	void bigPCFShadow(depthMapNode leftNode);
};
