#pragma once
#include <list>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include ".\Vector.h"
#include <string>
#include <vector>

class Renderer;
class Scene;
class GameManager;
class Quad;

#define SPTR boost::shared_ptr
#define WPTR boost::weak_ptr

/**
Node class
*/

class Node
{
public:
	friend class Scene;
	/**
	Constructor for node object
	*/
	Node(void);
	/**
	Destructor for node object
	*/
	~Node(void);

	/**
	Set the renderer for this node
	*/
	virtual void setRenderer(SPTR<Renderer> &_myRenderer);
	/**
	remove the renderer for this node
	*/
	virtual void removeRenderer();
	/**
	Returns the set renderer
	*/
	virtual SPTR<Renderer> getRenderer();
	/**
	Returns true if the Node has a renderer set
	*/
	bool hasRenderer();

	/** 
	enables Matrix Calculation
	*/
	void setEnableMatrixCalculation(bool _enableMatrixCalculation);
	/**
	Translate node
	*/
	virtual void translate(float dx, float dy, float dz);
	/**
	Translate node
	*/
	virtual void translate(Vector &tv);
	/**
	Rotate node
	*/
	virtual void rotate(float dalpha, float dbeta, float dgamma);
	/**
	Set node position
	*/
	virtual void setPosition(float x, float y, float z);
	/**
	Set node position
	*/
	virtual void setPosition(Vector &p);
	/**
	Set node rotation
	*/
	virtual void setRotation(float alpha, float beta, float gamma);
	/**
	Set node rotation
	*/
	virtual void setRotation(Vector &r);
	/**
	Returns node position
	*/
	Vector getPosition();
	/**
	Returns node X rotation
	*/
	float getRotationX();
	/**
	Returns node Y rotation
	*/
	float getRotationY();
	/**
	Returns node Z rotation
	*/
	float getRotationZ();

	/**
	Adds a child node
	*/
	std::list<SPTR<Node> >::iterator addChild(SPTR<Node> _child);
	/** 
	Removes a child node
	*/
	virtual void removeChild(std::list<SPTR<Node> >::iterator _it);
	/** 
	Removes a child node
	*/
	virtual void removeChild(SPTR<Node> _child);
	/** 
	Removes all child nodes
	*/
	virtual void removeAllChildren();
	/** 
	Kills a node
	*/
	virtual void killMe();

	/**
	update the node
	*/
	virtual void update(float dt);
	/**
	calculate world matrix for this node
	*/
	virtual void calcWorldMatrix(D3DXMATRIX &pMatWorld);
	/**
	Set world matrix for this node
	*/
	void setWorldMatrix(D3DXMATRIXA16 &worldMatrix);
	/**
	Returns the world matrix for this node
	*/
	D3DXMATRIXA16* getWorldMatrix();

	/**
	Returns the absolute vector for this node
	*/
	virtual void getAbsoluteVector(Vector &pOut, Vector &pin);
	/**
	Returns the absolute direction vector for this node
	*/
	virtual void getAbsoluteDirectionVector(Vector &dirOut, Vector &dirIn);
	/**
	Returns the absolute position vector for this node
	*/
	Vector getAbsolutePosition();

	/**
	Returns the absolute coordinates of the corner with the lowest values of the bounding box
	*/
	Vector* getAbsMinBBox();
	/**
	Returns the absolute coordinates of the corner with the highest values of the bounding box
	*/
	Vector* getAbsMaxBBox();
	/**
	Returns the absolute coordinates of the center of the bounding box
	*/
	Vector getAbsBoxCenter();
	/**
	Update the bounding box
	*/
	virtual void updateBBox();
	
	/**
	Set the scene for this node
	*/
	void setScene(Scene &scene);
	/**
	Set the visibility value for this node
	*/
	void setVisible(bool _visible);
	/**
	Returns whether the node is visible or not
	*/
	bool isVisible();
	/**
	Sets the node on standy node
	*/
	void setStandBy(bool _standBy);
	/**
	Returns whether the node is on standby or not
	*/
	bool onStandBy();

	/**
	The scene the node belongs to
	*/
	Scene *myScene;
	/**
	position vector for this node
	*/
	Vector myPosition;
	/**
	rotation vector for this node
	*/
	Vector myRotation;

	/**
	Returns the father of this node
	*/
	Node* getFather();
	/**
	Returns a list with cildren of this node
	*/
	std::list<SPTR<Node> > * getChildren();

	//Physic Stuff
	static const int NORMAL		=	0;
		
	//Particle Stuff
	virtual void setParticleGroup(int _group);
	int getParticleGroup();
	virtual Node* clone();
	virtual void setTimeToLive(float _timeToLive);
	float getTimeToLive();

	virtual void cloneChildren(Node *clonedNode);

	void orientAndMove(Vector direction, Vector upVec, Vector moveVec, bool moveRelative = true);

	Vector minAABBox;
	Vector maxAABBox;
	Vector absMinAABBox;
	Vector absMaxAABBox;
	Vector absBoxCenter;

	virtual void OnLostDevice( void* pUserContext );
	virtual void OnDestroyDevice( void* pUserContext );
	virtual HRESULT OnResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );

	/**
	defines whether this node should be frustum culled
	*/
	bool testAgainstFrustum;

	/**
	Returns whether this node is of the given type
	*/
	bool isA(int _nodeType);
	
	bool doingSoftKill;
	float softKillAlpha;
	virtual void setSoftKill(bool _softKill, float _duration = 1.0f);
	bool getSoftKill();
	float camDistance;

	friend class Quad;
	friend class ResourceManager;
protected:
	int nodeType;
	//Jede Node die auf StandBy=true ist, wird berhaupt nicht beachtet. Aktivierung durch explizites setzen oder Trigger(siehe goodie)
	bool standBy;
	//Soll updateMethode aufgerufen werden?
	bool updateMe;
	//Soll Node gezeichnet werden?
	bool visible;
	//Ist die vorhandene Node ein clon?
	bool aClone;
	//Calc Boundingbox of this Node
	bool calcBoundingBox;

	bool enableMatrixCalculation;

	bool softKill;
	float softKillDuration;
	
	Node *father;
	int behaveAs;
	std::list<SPTR<Node> > children;

	bool rendererAdded;
	SPTR<Renderer> myRenderer;

	D3DXMATRIXA16 myWorldMatrix;

	bool bOverrideWorldMatrix;
	void calcNormalWorldMatrix(D3DXMATRIX &pMatWorld);

	//Particle Stuff
	int particleGroup;
	float timeToLive;
	float maxAge;

	Vector myAbsPos;
	Vector corner[8];	//Fuer BoundingBox
	
	bool isZeroMatrix(const D3DMATRIX* pM);
};
