#pragma once
#include "Node.h"
#include "Quad.h"
#include <boost/shared_ptr.hpp>
#include <algorithm>
#include "Renderer.h"
#include "ResourceManager.h"
#include "RenderTarget.h"
#include "Camera.h"
#include "SimpleMeshRenderer.h"
#include "SkyBoxRenderer.h"
#include "Trigger.h"
#include "RenderPass.h"
#include "LowMemShadowPass.h"
#include "SharedResource.h"
#include "SharedResourceTexture.h"
#include "SoftShadowLightRenderer.h"

class Plane;
class GameManager;
class Object3d;
class soundNode;
class ResourceManager;
class Box;

#define NOFADE 0
#define FADEOUT 1
#define FADEIN 2

#define QUADLEVEL 5

#define SPTR boost::shared_ptr
#define WPTR boost::weak_ptr

/**
Scene class
*/

class Scene
{
public:
	
	friend class Node;
	friend class RenderPass;
	/**
	Constructor for the Scene object
	*/
	Scene(void);
	/**
	Destructor for the Scene object
	*/
	~Scene(void);

	/**
	Scene initialization
	*/
	virtual void initScene(GameManager &_manager);
	
	/**
	Returns root node
	*/
	Node* getRoot();

	/**
	Set scene visible/invisible
	*/
	void setVisible(bool _visible);
	/**
	Returns whether scene is visible or not
	*/
	bool isVisible();
	/**
	Fade in scene
	*/
	void fadeIn(float _duration);
	/**
	Fade out scene
	*/
	void fadeOut(float _duration);
	/**
	Set scene alpha
	*/
	void setSceneAlpha(float _alpha);
	
	/**
	Handle key press events
	*/
	void setKeyPressed(UINT key, bool bKeyPressed, bool bAltDown);
	/**
	Handle mouse events
	*/
	void setMouseStatus(bool bLeftButtonDown,  bool bRightButtonDown,bool bMiddleButtonDown,
                     bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta,
                     int xPos, int yPos);
	/**
	Returns mouse position
	*/
	int* getMousePos();
	/**
	Handle GUI events
	*/
	virtual void OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );

	/**
	Set scene width
	*/
	void setWidth(float _w);
	/**
	Set scene height
	*/
	void setHeight(float _h);
	/**
	Set scene depth 
	*/
	void setDepth(float _d);
	/**
	Returns Scene width
	*/
	float getWidth();
	/**
	Returns Scene height
	*/
	float getHeight();
	/**
	Returns Scene depth
	*/
	float getDepth();
	/**
	Clears scene data
	*/
	virtual void clearScene();

	/**
	Returns active Camera
	*/
	Camera * getActiveCamera();
	
	/**
	Render the scene
	*/
	virtual void renderScene(float fElapsedTime);
	
	struct renderSortFunction {
		bool operator()(Renderer* &r1, Renderer* &r2) {
			return r1->isLowerThan(r2);
		}
	};

	struct triggerSortFunction {
		bool operator()(SPTR<Trigger> &t1, SPTR<Trigger> &t2) {
			return t1->getTime() < t2->getTime();
		}
	};

	/**
	Returns active render pass
	*/
	RenderPass* getActiveRenderPass();

	/**
	Get delta time (time since last frame)
	*/
	float getDeltaTime();
	/**
	The dierctX device
	*/
	IDirect3DDevice9 * device;

	//Factory Stuff
	static const int NODE_OBJECT3D					= 1;
	static const int NODE_CAMERA					= 2;
	static const int NODE_SOUND						= 4;
	static const int NODE_BOX						= 8;
	static const int NODE_PARTICLEGROUP				= 16;
	static const int NODE_PARTICLEEMITTER			= 32;
	static const int NODE_SPRITE					= 64;
	static const int NODE_ORBITCAMERA				= 128;
	static const int NODE_SOFTSHADOWLIGHT			= 512;
	static const int NODE_SOFTSHADOWLOWMEMLIGHT		= 256;
	static const int NODE_SKYBOX					= 1024;

	virtual Node* createNode(int type);
	virtual Node* createNode(int type, Node &father);
	virtual Node* createNode(int type, bool addDefaultRenderer);
	virtual Node* createNode(int type, Node &father, bool addDefaultRenderer);
	/**
	Create a node in this scene
	*/
	virtual Node* createNode(int type, Node &father, bool addDefaultRenderer, bool isReference);
	virtual Node* createReferenceNode(int type, bool addDefaultRenderer);
	/**
	Connect Node and Renderer
	*/
	void connectNodeAndRenderer(Node &node, SPTR<Renderer> &renderer);
	/**
	The Game Manager
	*/
	GameManager *manager;
	
	//Shared Resource
	std::vector<SPTR<SharedResource> > sharedResourceVector;
	static const int SR_TEXTURE_VIEWDEPTH		= 1;	//DepthTexture Resource
	static const int SR_TEXTURE_DISTORTION		= 2;	//DistortionMap Resource
	static const int SR_TEXTURE_SHADOWMAP		= 3;	//ShadowMap Resource
	virtual SharedResource* createSharedResource(int id);
	SharedResource* getSharedResource(int id);

	//Device Stuff
	virtual void OnLostDevice( void* pUserContext );
	virtual void OnDestroyDevice( void* pUserContext );
	virtual HRESULT OnResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );

	//Trigger Stuff
	static const int TRIGGER_UNSETSTANDBY	= 0;
	static const int TRIGGER_KILLNODE		= 1;
	static const int TRIGGER_SETDETAIL		= 2;
	
	void setTrigger(int type, float time);
	void setTrigger(int type, float time, std::string text);
	void setTrigger(int type, Node *node, float time);
	void setTrigger(int type, Node *node, Node* secondNode, float time);
	void setTrigger(int type, Node *node, Node* secondNode, Vector normal, float time);
	float getSceneLiveTime();

	//Effect Stuff
	/**
	Returns Sun direction
	*/
	Vector getSunDirection();
	/**
	Set Sun direction
	*/
	void setSunDirection(Vector &v);
	/**
	Returns Sun position
	*/
	Vector getSunPosition();
	/**
	Set Sun position
	*/
	void setSunPosition(Vector &v);
	/**
	Returns Sun lookAt target
	*/
	Vector getSunLookAtTarget();
	/**
	Set Sun lookAt target
	*/
	void setSunLookAtTarget(Vector &v);
	/**
	Calculate light view matrix
	*/
	bool calculateLightViewMatrix();

	/**
	Returns world matrix
	*/
	D3DXMATRIX getWorldMatrix();
	/**
	Returns view matrix
	*/
	D3DXMATRIX getViewMatrix();
	/**
	Returns projection matrix
	*/
	D3DXMATRIX getProjectionMatrix();
	/**
	Returns identity matrix
	*/
	D3DXMATRIX getIdentityMatrix();
	/**
	Returns light view matrix
	*/
	D3DXMATRIX getLightViewMatrix();
	/**
	Returns light view matrix of reference point
	*/
	D3DXMATRIX getLightReferenceViewMatrix();
	
	/**
	Render scene nodes
	*/
	void renderNodes();
	Renderer* specialRenderer;

	/**
	Delete specified Node
	*/
	void deleteNode(Node* node);

	/**
	Returns Smart Pointer Of given Node
	*/
	SPTR<Node> getSmartPointer(Node *node);
	Object3d *sObj;
	/**
	Returns active render pass
	*/
	RenderPass *activeRenderPass;	

	// light:
	/**
	Set light direction 
	*/
	void setLight(Vector direction);
	/** 
	Set light view matrix
	*/
	void setLightViewMatrix(D3DXMATRIX _lightViewMatrix);
	/**
	Set light view matrix of reference point
	*/
	void setLightReferenceViewMatrix(D3DXMATRIX _lightRefViewMatrix);

	/**
	Returns SoftShadowLight List (Smart Pointer)
	*/
	std::list<SPTR<Node> >* getSoftShadowLightList();
	/**
	Returns current light
	*/
	LowMemLight* getCurrentLight();

	/**
	Renderer list
	*/
	std::vector<SPTR<Renderer> > rendererList;	//List of all renderer in scene

	/**
	Delete Specified Renderer in list
	*/
	void deleteRendererInList(Renderer* renderer);
	
	/**
	Defines whether bounding boxes should be drawn
	*/
	bool drawBBoxes;

	/**
	Render the final image
	*/
	void renderFinalImage(ID3DXEffect* finalEffect);
	/**
	Returns the active pass ID
	*/
	int getActivePassId();
	
	//Pass stuff
	static const int PASS_NORMAL = 0;
	static const int PASS_DEPTH = 2;
	static const int PASS_SHADOWMAP = 3;
	static const int PASS_SOFTSHADOWMAP = 3;
	static const int PASS_REPROJECTION = 4;
	static const int PASS_ADDSHADOWMAP = 5;
	static const int PASS_SOFTSHADOW = 6;
	
	/**
	set active render pass
	*/
	void setActiveRenderPass(RenderPass &_activePass);

	LPD3DXMESH getQuadMesh();
	
	FSOUND_STREAM *bgMusicStream;
	int bgSoundChannel;
	float bgSoundVolume;
	bool soundStopped;
	void setBackgroundSound(std::string filename);
	void setBackgroundSoundVolume(float vol);
	void playBackgroundSound();
	virtual void updateSoundNodeVolume();
	virtual void setBackgroundSoundMute(bool _playMusic);
	bool getBackgroundSoundMute();

	SharedResourceTexture* shadowMapTextureResource;
	LPDIRECT3DSURFACE9 shadowMapSurface;

	//Keyboard & Mouse Stuff
	bool keyDown[256];
	static const int MOUSE_LEFT		= 0;
	static const int MOUSE_RIGHT	= 1;
	static const int MOUSE_MIDDLE	= 2;
	static const int MOUSE_SIDE1	= 3;
	static const int MOUSE_SIDE2	= 4;
	bool mouseButton[5];
	int  mousePos[2];
	int  mouseWheel;
	bool altDown;

	/**
	Clear Render Target
	*/
	void clearRenderTargetAndZBuffer();
	
	IDirect3DTexture9* getViewDepthMap();

	IDirect3DTexture9* finalImage;

	//Diverse Rendering Modes
	int sumMode;

	static const int SUM_ONEPASS		= 0;
	static const int SUM_WHOLESCENE		= 1;
	static const int SUM_TWOPASS		= 2;

	/**
	Filter mode: None, PCF, PCF_1D
	*/
	int filterMode;

	static const int FILTER_NONE		= 0;
	static const int FILTER_PCF			= 1;
	static const int FILTER_PCF_1D		= 2;

	/**
	View Point: Camero or Light Source
	*/
	int viewPointMode;

	static const int VIEWPOINT_LIGHT	= 0;
	static const int VIEWPOINT_CAMERA	= 1;
	
	/**
	Shadowing mode: our algorithm, 128 maps, big pcf
	*/
	int shadowingMode;

	static const int SHMODE_NORMAL		= 0;
	static const int SHMODE_SIMPLE		= 1;
	static const int SHMODE_BIGPCF		= 2;
	
	/**
	Indicates whether debug information is shown
	*/
	bool showDebugInformation;
	/**
	Indicates whether textures are shown
	*/
	bool useTextures;
protected:
	//Scene Management Stuff
	bool visible;
	float fadeDuration;
	float fadeTimer;
	int fadeState;
	float sceneAlpha;
	bool muteMusic;
	Box* boundingBox;

	bool showSoftShadowMaps;
	LowMemLight* currentLight;
	void renderShowShadowMaps();
	
	float width;
	float height;
	float depth;

	//RenderChain Stuff
	std::list<void (Scene::*)(void)> renderChain;
	bool firstFrame;

	void takeTime();
	void doFade();
	void updateTrigger();
	void updateNodes();
	void calculateTransformations();
	void doFrustumCulling();
	void doViewingTransformation();
	virtual void cleanUpScene();
	void startRenderPasses();
	void doFinalSoftShadowPass();

	//Clear operations
	void clearSharedResources();
	
	float dt;
	LARGE_INTEGER frequency;
	LARGE_INTEGER lastFrame;
	LARGE_INTEGER currentFrame;

	//Culling Stuff
	Quad fcQuad;
	void buildQuadTree();
	bool enableFrustumCulling;
	bool quadTreeBuilt;

	//Scene Objects Stuff
	Node root;
	D3DXMATRIX worldMatrix;
	D3DXMATRIX identMatrix;
	
	D3DXMATRIX lightViewMatrix;
	D3DXMATRIX lightRefViewMatrix;
	
	std::list<SPTR<Node> > visibleNodeList;
	std::list<SPTR<Node> > zombieNodeList;
	Camera *activeCamera;
	Camera *defaultCamera;

	std::list<SPTR<Node> > nodeList;			//List of all nodes in scene
	
	std::vector<SPTR<Node> > refObject3dList;	//List of all object3d s in scene
	std::vector<SPTR<Node> > refCameraList;		//List of all cameras
	std::vector<SPTR<Node> > refSoundList;		//List of soundNodes
	
	std::list<SPTR<Node> > soundNodeList;		//List of all SoundNodes in Scene
	std::list<SPTR<Node> > softShadowLightList;	//List of all SoftShadowLights in Scene
	void deleteNodeInList(Node* node, std::list<SPTR<Node> > &list);

	//Sun Stuff
	Vector sunPosition;
	Vector sunLookAtTarget;
	
	Vector sunDirection;
	D3DLIGHT9 d3dLight;

	//TriggerStuff
	float sceneLifeTime;
	std::vector<SPTR<Trigger> >	triggerList;
	virtual void executeTrigger(SPTR<Trigger> trigger);

	//Fading & Postprocessing Stuff
	IDirect3DSurface9* finalImageSurface;
	IDirect3DTexture9* reprojectionTexture;

	D3DSURFACE_DESC finalImageDesc;
	D3DXMATRIX spriteMatrix;
	LPD3DXMESH quadMesh;
	bool meshCreated;
	void createQuadMesh();

	struct Vertex
	{
		Vertex(float _x, float _y, float _z, float _u, float _v)
		{
			x = _x;  
			y = _y;   
			z = _z;
			u = _u;
			v = _v;
		}

		Vertex(Vector vec, float _u, float _v)
		{
			x = vec.x;  
			y = vec.y;   
			z = vec.z;
			u = _u;
			v = _v;
		}

		float x, y, z, u, v;
	};

	const static DWORD FVF_Flags = D3DFVF_XYZ | D3DFVF_TEX1;

	RenderPass defaultRenderPass;
	RenderPass depthRenderPass;
	RenderPass shadowMapRenderPass;
	LowMemShadowPass softShadowRenderPass;

	SharedResourceTexture* depthTextureResource;
	LPDIRECT3DSURFACE9 depthSurface;
	bool needDepthPass;
	bool needDistortionPass;
	bool needShadowMapPass;

	IDirect3DTexture9* distortionMap1;
	IDirect3DTexture9* distortionMap2;
	IDirect3DTexture9* preDistortionFinal;
	IDirect3DSurface9* preDistortionFinalSurface;
	IDirect3DTexture9* screenDistortionMap;
	IDirect3DSurface9* screenDistortionMapSurface;
};
