#include "dxstdafx.h"
#include ".\simplemeshrenderer.h"
#include "GameManager.h"
#include "LowMemLight.h"

SimpleMeshRenderer::SimpleMeshRenderer(void) : Renderer()
{
	this->depthMapEffect = NULL;
	this->simpleMeshEffect = NULL;
	this->alphaBlending = false;
	this->setRenderPriority(30);
	this->rendererType |= Renderer::RENDERER_SIMPLEMESH;
}

SimpleMeshRenderer::~SimpleMeshRenderer(void)
{
	this->myScene->manager->releaseEffect(this->simpleMeshEffect);
	this->myScene->manager->releaseEffect(this->depthMapEffect);
	this->depthMapEffect = NULL;
	this->simpleMeshEffect = NULL;
}


void SimpleMeshRenderer::init()
{
	this->simpleMeshEffect = this->myScene->manager->getEffect(GameManager::EFFECT_SIMPLEMESH);
	if(!this->simpleMeshEffect) {
		this->simpleMeshEffect = this->myScene->manager->loadEffect(GameManager::EFFECT_SIMPLEMESH, L"shaders/simpleMesh.fx");
	}
	this->depthMapEffect = this->myScene->manager->getEffect(GameManager::EFFECT_DEPTHMAP);
	if(!this->depthMapEffect) {
		this->depthMapEffect = this->myScene->manager->loadEffect(GameManager::EFFECT_DEPTHMAP, L"shaders/depthPass.fx");
	}
}


void SimpleMeshRenderer::render() 
{
	this->device->SetTransform(D3DTS_WORLD, &this->myScene->getIdentityMatrix());
	this->device->SetTransform(D3DTS_VIEW, &this->myScene->getIdentityMatrix());
	this->device->SetTransform(D3DTS_PROJECTION, &this->myScene->getIdentityMatrix());
	switch(this->myScene->getActivePassId()) {
		case Scene::PASS_NORMAL:
			this->defaultPass();
			break;
		case Scene::PASS_DEPTH:
			this->depthPass();
			break;
		case Scene::PASS_SHADOWMAP:
		case Scene::PASS_REPROJECTION:
			this->shadowMapPass();
			break;
		case Scene::PASS_ADDSHADOWMAP:
			this->addSMPass();
			break;
	}
}

void SimpleMeshRenderer::defaultPass()
{
	HRESULT hr = 0;
	SPTR<Node> node = this->myNode.lock();
	if(node->isA(Scene::NODE_SPRITE)) {
		return;
	}
		
	Object3d *obj = (Object3d *) (node.get());
	if(obj!=NULL && obj->isModelLoaded()) {
		UINT cPass;
		this->device->BeginScene();
		
		simpleMeshEffect = this->myScene->manager->getEffect(GameManager::EFFECT_SIMPLEMESH);
		
		V( simpleMeshEffect->SetFloat("alpha", obj->softKillAlpha) );
		V( simpleMeshEffect->SetMatrix("g_mWorldViewProjection", &this->myScene->activeRenderPass->getWorldViewProjectionMatrix(obj->getWorldMatrix())) );
		
		float nearPlane = this->myScene->getActiveCamera()->getNearClipping();
		float farPlane = this->myScene->getActiveCamera()->getFarClipping();
		
		V( simpleMeshEffect->SetFloat("FarPlaneMinusNearPlane", farPlane - nearPlane) );

		D3DXMATRIX modelViewIT;
		modelViewIT = *obj->getWorldMatrix();
		D3DXMATRIX trans;
		Vector absPos = obj->getAbsolutePosition();
		D3DXMatrixTranslation(&trans, -absPos.x, -absPos.y, -absPos.z);
		D3DXMatrixMultiply(&modelViewIT, &modelViewIT, &trans);
		D3DXMatrixTranspose(&modelViewIT, &modelViewIT);
		
		Vector lightDir;
		Vector eyeDir;
		Vector halfVector;
		lightDir = ((this->myScene->getSunDirection()) * -1);
		eyeDir = this->myScene->getActiveCamera()->getAbsolutePosition() - obj->getAbsolutePosition();
		halfVector = lightDir + eyeDir;
		halfVector.normalize();
		lightDir.normalize();
		
		V( simpleMeshEffect->SetMatrix("g_mWorldViewIT", &modelViewIT) );
		V( simpleMeshEffect->SetVector("lightDir", &lightDir));
		V( simpleMeshEffect->SetVector("halfVector", &halfVector) );
		V( simpleMeshEffect->CommitChanges() );
		
		int iNumPasses = 0;

		if(myScene->useTextures) {
			V( simpleMeshEffect->SetTechnique("simpleMesh") );
		} else {
			V( simpleMeshEffect->SetTechnique("simpleMeshNT") );
		}
		
		iNumPasses = 1;
						
		V( simpleMeshEffect->Begin( &cPass, 0 ) );
		
		if(obj->doingSoftKill) {	
			this->device->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
			this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
			this->device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
			this->device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
		}
		if (obj->Materials.size() == 0 || !myScene->useTextures) {
			if (obj->Materials.size() != 0) {
				V( simpleMeshEffect->SetFloatArray("ambient", (float*) (&obj->Materials[0].Ambient), 4) );
				V( simpleMeshEffect->SetFloatArray("diffuse", (float*) (&obj->Materials[0].Diffuse), 4) );
				V( simpleMeshEffect->SetFloatArray("specular", (float*) (&obj->Materials[0].Specular), 4) );
				V( simpleMeshEffect->SetFloat("power", obj->Materials[0].Power) );
			}
			V( simpleMeshEffect->CommitChanges() );
			this->device->SetTexture(0, 0);
			V( simpleMeshEffect->BeginPass( 0 ) );

			this->device->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
			this->device->SetRenderState( D3DRS_LIGHTING, FALSE );
			this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

			for(UINT i = 0; i < obj->Materials.size(); i++) {
				(*obj->getMesh())->DrawSubset(i);
			}
			this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
			
			V( simpleMeshEffect->EndPass() );
		} else {
			this->device->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW);
			this->device->SetRenderState( D3DRS_LIGHTING, FALSE );
			this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
			
			// bilinear texture filtering:
			this->device->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
			this->device->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
			this->device->SetSamplerState( 0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
			
			for(int i = 0; i < iNumPasses; i++) {

				V( simpleMeshEffect->BeginPass( i ) );

				for(UINT i = 0; i < obj->Materials.size(); i++) {

					V( simpleMeshEffect->SetFloatArray("ambient", (float*) (&obj->Materials[i].Ambient), 4) );
					V( simpleMeshEffect->SetFloatArray("diffuse", (float*) (&obj->Materials[i].Diffuse), 4) );
					V( simpleMeshEffect->SetFloatArray("specular", (float*) (&obj->Materials[i].Specular), 4) );
					V( simpleMeshEffect->SetFloat("power", obj->Materials[i].Power) );
					V( simpleMeshEffect->SetTexture("objTexture", obj->Textures[0]) );
					V( simpleMeshEffect->CommitChanges() );
					
					(*obj->getMesh())->DrawSubset(i);
				}
			
				V( simpleMeshEffect->EndPass() );

			}

			if(obj->doingSoftKill) {
				this->device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
				this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
			}
		}

		V( simpleMeshEffect->End() );

		this->device->SetTransform(D3DTS_WORLD, &this->myScene->getIdentityMatrix());
		this->device->SetTransform(D3DTS_VIEW, &this->myScene->getIdentityMatrix());
		this->device->SetTransform(D3DTS_PROJECTION, &this->myScene->getIdentityMatrix());
		this->device->EndScene();
	}
}

void SimpleMeshRenderer::depthPass()
{
	SPTR<Node> node = this->myNode.lock();
	Object3d *obj = (Object3d *) (node.get());
	UINT cPass;
	HRESULT hr;
	if(obj!=NULL && obj->isModelLoaded()) {
		V( this->depthMapEffect->SetMatrix("g_mWorldViewProjection", &this->myScene->activeRenderPass->getWorldViewProjectionMatrix(obj->getWorldMatrix())) );
		V( this->depthMapEffect->SetTechnique("DepthPass") );
		float nearPlane = this->myScene->getActiveCamera()->getNearClipping();
		float farPlane = this->myScene->getActiveCamera()->getFarClipping();
		V( this->depthMapEffect->SetFloat("FarPlaneMinusNearPlane", farPlane - nearPlane) );
		V( this->depthMapEffect->CommitChanges() );
		this->device->BeginScene();
		V( this->depthMapEffect->Begin( &cPass, 0 ) );
		V( this->depthMapEffect->BeginPass( 0 ) );
		if (obj->Materials.size() == 0) {
			this->device->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
			this->device->SetRenderState( D3DRS_LIGHTING, FALSE );
			this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
			
			(*obj->getMesh())->DrawSubset(0);

			this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
		} else {
			this->device->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW);
			this->device->SetRenderState( D3DRS_LIGHTING, TRUE );
			this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
			if(this->alphaBlending) {
				this->device->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
				this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
				this->device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
				this->device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
			}
			
			// bilinear texture filtering:
			this->device->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
			this->device->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
			this->device->SetSamplerState( 0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );

			for(UINT i = 0; i < obj->Materials.size(); i++) {
				this->device->SetMaterial( &obj->Materials[i] );
				this->device->SetTexture(0, obj->Textures[i]);
				
				(*obj->getMesh())->DrawSubset(i);
			}
			if(this->alphaBlending) {
				this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
				this->device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
			}	
		}
		V( this->depthMapEffect->EndPass() );
		V( this->depthMapEffect->End() );
		this->device->EndScene();
	}
}

void SimpleMeshRenderer::shadowMapPass()
{
	SPTR<Node> node = this->myNode.lock();
	Object3d *obj = (Object3d *) (node.get());
	UINT cPass;
	HRESULT hr;
	if(obj!=NULL && obj->isModelLoaded()) {
		
		// get shadow map effect:
		ID3DXEffect* shadowMapEffect;
		shadowMapEffect = this->myScene->manager->getEffect(GameManager::EFFECT_SHADOWMAP);

		D3DXMATRIX worldViewMat;
		D3DXMATRIX worldViewProjMat;
		D3DXMATRIX worldRefViewMat;
		D3DXMATRIX worldRefViewProjMat;
	
		// set the technique to use
		if(this->myScene->getActivePassId()!=Scene::PASS_REPROJECTION) {
			shadowMapEffect->SetTechnique( "ShadowMap" );
		} else {
			shadowMapEffect->SetTechnique("Reprojection");
				
			D3DXMatrixMultiply(&worldRefViewMat, obj->getWorldMatrix(), &myScene->getLightReferenceViewMatrix());
			D3DXMatrixMultiply(&worldRefViewProjMat, &worldRefViewMat, &myScene->getProjectionMatrix());

			V( shadowMapEffect->SetMatrix("refProjection", &worldRefViewProjMat) );

			float invHSW = 1.0f/(2.0f*myScene->manager->screenWidth);
			float invHSH = 1.0f/(2.0f*myScene->manager->screenHeight);
			V( simpleMeshEffect->SetFloat("invHalfScreenWidth", invHSW) );
			V( simpleMeshEffect->SetFloat("invHalfScreenHeight", invHSH) );
		}
		
		D3DXMatrixMultiply(&worldViewMat, obj->getWorldMatrix(), &myScene->getLightViewMatrix());
		D3DXMatrixMultiply(&worldViewProjMat, &worldViewMat, &myScene->getProjectionMatrix());

		V( shadowMapEffect->SetMatrix("LightWorldViewProjection", &worldViewProjMat) );

		float nearPlane = this->myScene->getActiveCamera()->getNearClipping();
		float farPlane = this->myScene->getActiveCamera()->getFarClipping();
		shadowMapEffect->SetFloat("FarPlaneMinusNearPlane", farPlane - nearPlane);

		//get light (used to create light matrices)
		LowMemLight* tempLight = myScene->getCurrentLight();
		V( simpleMeshEffect->SetFloat("SMBias", tempLight->getSMBias()) );
		
		shadowMapEffect->CommitChanges();
		
		V( shadowMapEffect->Begin(&cPass, 0) );
		for(UINT j = 0; j < cPass; j++)	{
			V( shadowMapEffect->BeginPass( j ) );

			this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
			if(this->alphaBlending) {
				this->device->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
				this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
				this->device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
				this->device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
			}
			
			for(UINT i = 0; i < obj->Materials.size(); i++) {
				(*obj->getMesh())->DrawSubset(i);
			}
			if(this->alphaBlending) {
				this->device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
				this->device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
			}

			V( shadowMapEffect->EndPass() );
		}

		V( shadowMapEffect->End() );
	}
}

void SimpleMeshRenderer::addSMPass()
{
	SPTR<Node> node = this->myNode.lock();
	Object3d *obj = (Object3d *) (node.get());
	UINT cPass;
	HRESULT hr;
	if(obj!=NULL && obj->isModelLoaded()) {
		//get effect
		simpleMeshEffect = this->myScene->manager->getEffect(GameManager::EFFECT_SIMPLEMESH);
		
		//set technique
		if(this->myScene->shadowingMode == Scene::SHMODE_BIGPCF) {
			V( simpleMeshEffect->SetTechnique("sumShadowsPCF") );
			//set values:
			V( simpleMeshEffect->SetFloat("samplingDistance", myScene->getCurrentLight()->getFilterSamplingDistance()) );
			V( simpleMeshEffect->SetInt("kernelSize", (int) myScene->getCurrentLight()->getLightDimension()) );
		} else if(this->myScene->filterMode == this->myScene->FILTER_PCF) {
			switch (myScene->sumMode) {
				case Scene::SUM_ONEPASS:
					V( simpleMeshEffect->SetTechnique("sumShadowsPCF") );
					//set values:
					V( simpleMeshEffect->SetFloat("samplingDistance", myScene->getCurrentLight()->getFilterSamplingDistance()) );
					V( simpleMeshEffect->SetInt("kernelSize", myScene->getCurrentLight()->getFilterKernelSize()) );
					break;
				case Scene::SUM_WHOLESCENE:
					V( simpleMeshEffect->SetTechnique("sumAdvancedShadows") );
					break;
				case Scene::SUM_TWOPASS:
					V( simpleMeshEffect->SetTechnique("singleShadowOnly") );
					break;
			}
		} else if(this->myScene->filterMode == this->myScene->FILTER_PCF_1D) {
			switch (myScene->sumMode) {
				case Scene::SUM_ONEPASS:
					V( simpleMeshEffect->SetTechnique("sumShadowsPCF1D") );
					//set values:
					V( simpleMeshEffect->SetFloat("samplingDistance", myScene->getCurrentLight()->getFilterSamplingDistance()) );
					V( simpleMeshEffect->SetInt("kernelSize", myScene->getCurrentLight()->getFilterKernelSize()) );
					break;
				case Scene::SUM_WHOLESCENE:
					V( simpleMeshEffect->SetTechnique("sumAdvancedShadows") );
					break;
				case Scene::SUM_TWOPASS:
					V( simpleMeshEffect->SetTechnique("singleShadowOnly") );
					break;
			}
		} else {
			switch (myScene->sumMode) {
				case Scene::SUM_ONEPASS:
					V( simpleMeshEffect->SetTechnique("sumShadows") );
					break;
				case Scene::SUM_WHOLESCENE:
					V( simpleMeshEffect->SetTechnique("sumAdvancedShadows") );
					break;
				case Scene::SUM_TWOPASS:
					V( simpleMeshEffect->SetTechnique("singleShadowOnly") );
					break;
			}
		}

		//set worldviewprojection matrix
		V( simpleMeshEffect->SetMatrix("g_mWorldViewProjection", &this->myScene->activeRenderPass->getWorldViewProjectionMatrix(obj->getWorldMatrix())) );
		float invHSW = 1.0f/(2.0f*myScene->manager->screenWidth);
		float invHSH = 1.0f/(2.0f*myScene->manager->screenHeight);
		V( simpleMeshEffect->SetFloat("invHalfScreenWidth", invHSW) );
		V( simpleMeshEffect->SetFloat("invHalfScreenHeight", invHSH) );
		
		//set floats
		float nearPlane = this->myScene->getActiveCamera()->getNearClipping();
		float farPlane = this->myScene->getActiveCamera()->getFarClipping();
		V( simpleMeshEffect->SetFloat("FarPlaneMinusNearPlane", farPlane - nearPlane) );

		//get light (used to create light matrices)
		LowMemLight* tempLight = myScene->getCurrentLight();
		
		//set the render target as lookup texture
		V( simpleMeshEffect->SetTexture("renderTarget", tempLight->getLightSumTexture()) );

		//set the scene depth map for depth comparison:
		V( simpleMeshEffect->SetTexture("sceneDepth", myScene->getViewDepthMap()) );

		//create & set matrix:
		D3DXMATRIX worldLightMat;
		D3DXMATRIX worldLightProjMat;
		D3DXMatrixMultiply(&worldLightMat, obj->getWorldMatrix(), tempLight->getCurrentWorldMatrix());
		D3DXMatrixMultiply(&worldLightProjMat, &worldLightMat, &myScene->getProjectionMatrix());
		
		V( simpleMeshEffect->SetMatrix("g_mLightWorldViewProjection", &worldLightProjMat) );

		//set the texture
		V( simpleMeshEffect->SetTexture("ShadowMap", tempLight->getCurrentDepthMap()) );
		
		//commit all changes
		V( simpleMeshEffect->CommitChanges() );

		int iNumPasses = 1;
						
		V( simpleMeshEffect->Begin( &cPass, 0 ) );
		
		this->device->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW);
		this->device->SetRenderState( D3DRS_LIGHTING, FALSE );
		this->device->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
		
		for(int i = 0; i < iNumPasses; i++) {
			V( simpleMeshEffect->BeginPass( i ) );
			for(UINT i = 0; i < obj->Materials.size(); i++) {
				V( simpleMeshEffect->SetFloatArray("ambient", (float*) (&obj->Materials[i].Ambient), 4) );
				V( simpleMeshEffect->SetFloatArray("diffuse", (float*) (&obj->Materials[i].Diffuse), 4) );
				V( simpleMeshEffect->SetFloatArray("specular", (float*) (&obj->Materials[i].Specular), 4) );
				V( simpleMeshEffect->SetFloat("power", obj->Materials[i].Power) );
				V( simpleMeshEffect->SetTexture("objTexture", obj->Textures[i]) );
				V( simpleMeshEffect->CommitChanges() );
				
				(*obj->getMesh())->DrawSubset(i);
			}
			
			V( simpleMeshEffect->EndPass() );
		}

		V( simpleMeshEffect->End() );
	}
}

void SimpleMeshRenderer::enableAlphaBlending(bool _alpha)
{
	this->alphaBlending = _alpha;
}

void SimpleMeshRenderer::OnLostDevice( void* pUserContext )
{
	HRESULT hr;
	this->myScene->manager->printToConsole(" OnLostDevice SimpleMeshRenderer");
	V( this->simpleMeshEffect->OnLostDevice() );
	V( this->depthMapEffect->OnLostDevice() );
}

void SimpleMeshRenderer::OnDestroyDevice( void* pUserContext )
{
	this->myScene->manager->printToConsole(" OnDestroyDevice SimpleMeshRenderer");
	this->myScene->manager->releaseEffect(this->simpleMeshEffect);
	this->myScene->manager->releaseEffect(this->depthMapEffect);
	this->simpleMeshEffect = NULL;
	this->depthMapEffect = NULL;
}

HRESULT SimpleMeshRenderer::OnResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
	this->device = pd3dDevice;
	this->simpleMeshEffect = this->myScene->manager->getEffect(GameManager::EFFECT_SIMPLEMESH);
	if(!this->simpleMeshEffect) {
		this->simpleMeshEffect = this->myScene->manager->loadEffect(GameManager::EFFECT_SIMPLEMESH, L"shaders/simpleMesh.fx");
	}
	this->depthMapEffect = this->myScene->manager->getEffect(GameManager::EFFECT_DEPTHMAP);
	if(!this->depthMapEffect) {
		this->depthMapEffect = this->myScene->manager->loadEffect(GameManager::EFFECT_DEPTHMAP, L"shaders/depthPass.fx");
	}
	return NULL;
}


bool SimpleMeshRenderer::isLowerThan(Renderer* renderer)
{
	SPTR<Node> node1 = this->myNode.lock();
	Object3d *myObj = (Object3d *) (node1.get());
	SPTR<Node> node2 = renderer->myNode.lock();
	Object3d *otherObj = (Object3d *) (node2.get());
	if(this->getRenderPriority()<renderer->getRenderPriority()) {
		return true;
	} else {
		if(this->getRenderPriority()==renderer->getRenderPriority()) {
			if(myObj->doingSoftKill && !otherObj->doingSoftKill) {
				return false;
			}
			if(!myObj->doingSoftKill && otherObj->doingSoftKill) {
				return true;
			}
			return myObj->camDistance>otherObj->camDistance;
		} else {
			return false;
		}
	}
}

